%!octave 
%      
%
%
%       OS : GNU/Linux 4.2.5
% COMPILER : GNU Octave, version 4.0.0
%
%   AUTHOR : Evgeny S. Borisov
%
%  Glushkov Institute of Cybernetics
%  National Academy of Sciences of Ukraine
% 
%    http://www.mechanoid.kiev.ua
%  e-mail : nn@mechanoid.kiev.ua
% 
% $Id: mlp_backprop_grad.m,v 1.2 2016/04/12 13:19:18 mechanoid Exp $


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% градиент функции ошибки сети
%
% L - учебный набор 
% W - матрица весов
% cfg - описание сети
% dlm - маска dropout для слоёв
%
function retval=mlp_backprop_grad(L,W,cfg),
   ltp=2; % номер ячейки cfg с типом ф-ции активации

   % фаза прямого распространения
   l=mlp_layers( L{1}, W, cfg ); % состояния нейронов всех слоёв

   % фаза обратного распространения
   er=mlp_errors(l,L{2},W,cfg); % ошибки сети на учебном наборе

   g=mlp_grad(l,er,cfg); % градиент ф-ции ошибки сети

   %g=mlp_weight_norm(g); 

   retval=g;
endfunction

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval=mlp_grad(l,er,cfg),
   ltp = 2 ; % номер ячейки cfg с типом ф-ции активации
   g = {} ; % градиент ошибки сети
   
   for j=1:length(er),
      o = mlp_activate(l{j},cfg{j}{ltp})  ;
	  g{j}{1} = er{j}*o' ; 
	  g{j}{2} = sum(er{j},2) ;
   endfor


   retval=g;

endfunction

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval=mlp_errors(l,y,W,cfg),
   ltp=2; % номер ячейки cfg с типом ф-ции активации
   layers_numb = length(cfg) ; % количество слоёв сети, включая вход
 
   er = {} ; % ошибка слоёв
  
   er{layers_numb-1} = mlp_loss_drvt( l{end}, y, cfg ) ; % ошибка сети (выходного слоя)
  
   for j=layers_numb-2:-1:1, % ошибки скрытых слоёв
	  df = mlp_activate_drvt( l{j+1}, cfg{j+1}{ltp} ) ; % значения производной ф-ции активации 
	  er{j} = ( (W{j+1}{1})' * er{j+1} ) .* df ; % ошибка слоя j
   endfor

   retval=er;
endfunction

