%!octave 
%
%
%       OS : GNU/Linux 4.2.5-1-ARCH 
% COMPILER : GNU Octave, version 4.0.0
%
%   AUTHOR : Evgeny S. Borisov
%
%  Glushkov Institute of Cybernetics
%  National Academy of Sciences of Ukraine
% 
%    http://www.mechanoid.kiev.ua
%  e-mail : nn@mechanoid.kiev.ua
% 
% $Id: mlp_trn_adam.m,v 1.4 2016/04/21 18:23:47 mechanoid Exp mechanoid $


% REFERENCES:
%
% ADAM: A METHOD FOR STOCHASTIC OPTIMIZATION
% 
% Diederik P. Kingma
% University of Amsterdam
% dpkingma@uva.nl
% 
% Jimmy Lei Ba
% University of Toronto
% jimmy@psi.utoronto.ca


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% обучение ИНС методом градиентного спуска (Adam)
%
% L={Xl,Yl} - учебный набор
% С={Xc,Yc} - контрольный набор
% W - начальная матрица весов
% cfg - описание параметров сети сети
% trn_par - параметры обучения
%
function [W,er_hist]=mlp_trn_adam(L,C,W,cfg,trn_par),
   ltp = 2 ; % номер ячейки cfg с типом ф-ции активации
   min_err = trn_par{1} ; % порог допустимой ошибки обучения
   max_iter = trn_par{2} ; % максимальное число эпох
   er_win_size = trn_par{3} ;  % размер окна контроля за ошибкой
   trsp = trn_par{4} ; % параметр скорости обучения
   moment_par = trn_par{5} ; % параметр момента 
   reg_par = trn_par{6} ; % параметр регуляризации
   batch_par = trn_par{7} ; % доля учебных примеров в одном пакете от общего количества примеров
   print_step = trn_par{8} ;% печатаем значение ошибки каждые print_step эпох 

   X=L{1} ; y=L{2} ; % учебный набор
   num_examples=size(y,2) ; % количество объектов в учебном наборе
   batch_size = max(2,round(num_examples*batch_par)) ; % количество учебных примеров в одном пакете 

   tname="Adam";
   printf("[i] обучение методом %s\n",tname);
   save("-text","result/tname.txt", "tname" ) ;

   if batch_size < num_examples,
	  printf("[i] размер пакета (mini-batch) %i / %i \n",batch_size,num_examples);
   else
	  printf("[i] размер учебного набора (full-batch) %i \n",num_examples);
   endif

   er_hist = [er_win_size:-1:1]*100.0 ; % история значений ошибки обучения

   dW = m = m2 = mlp_weight_set(cfg,0.0) ; % начальные значения моментов

   % начальные значения суммы квадратов градиента
   for i=1:length(W), sg{i}{1} = sg{i}{2} = 0.0 ; endfor

   for iter=1:max_iter,

	  er_hist(end+1) = mlp_loss(C,W,cfg,reg_par); % вычисляем ошибку сети на контрольном наборе

	  if mod(iter,print_step)<1, % печатаем значение ошибки print_step эпох 
		 printf("      %i/%i : %f / %f   , %f     \r",iter,max_iter,er_hist(end),min_err,trsp ) ; 
      endif
 	 
	  if trn_stop_condition(er_hist,min_err,er_win_size), break ; endif % условия остановки обучения
 
	  % случайный пакет из учебных примеров
      batch_idx = randperm(num_examples)(1:batch_size); 
	  Lr = { X(:,batch_idx) , y(:,batch_idx) };
      
	  g = mlp_backprop_grad(Lr,W,cfg); % значение градиента
	  %g = mlp_weight_norm(g); 

	  b=0.9 ; a=0.999 ;

	  m = acc_weigth(g,m,b);
	  m2 = acc_weigth2(g,m2,a);

	  gn = grad_norm(m,m2,a,b); % изменение весов 

	  dW = weight_change(gn,dW,W,trsp,reg_par,moment_par); % полное изменение весов с регуляризацией
	  % dW = weight_change(gn,W,trsp,reg_par); 

	  W = mlp_weight_add(W,dW,-1.0); % корректируем веса 
	 
   endfor

   printf("[i]   %i/%i : %f / %f   , %f     \n",iter,max_iter,er_hist(end),min_err,trsp ) ; 

   er_hist(1:er_win_size)=[]; % удаляем начальные значения ошибки из истории 

   if iter>(max_iter-1) , printf("[w] достигнут порог количества циклов обучения         \n") ; endif

endfunction



% b=0.9 ; a=0.999 ; c=1e-7 ;
% m = v = 0;
% m := b*m + (1-b)*g
% v := a*v + (1-a)*(g^2)
% dw = trsp * ( m/(1-b) ) /(sqrt(v/(1-a))+c)


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval = acc_weigth2(w,s,a),
   r={};
   for i=1:length(w),
	  for j=1:2,
		 r{i}{j} = a*s{i}{j} + (1-a)*( (w{i}{j}).^2 ) ;
	  endfor
   endfor
 
   retval=r;
endfunction
% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval = acc_weigth(w,s,a),
   r={};
   for i=1:length(w),
	  for j=1:2,
		 r{i}{j} = a*s{i}{j} + (1-a)*w{i}{j} ;
	  endfor
   endfor
 
   retval=r;
endfunction



% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval = grad_norm(m,m2,a,b), 
   c=1e-7 ;

   d = sqrt(1-a)/(1-b) ;

   r={};
   for i=1:length(m),
	  for j=1:2,
		 r{i}{j} = m{i}{j} ./ ( sqrt( m2{i}{j} ) + c ) * d ;
		 %r{i}{j} = ( m{i}{j}/(1-b) ) ./ ( sqrt( m2{i}{j}./(1-a) ) + c ) ;
	  endfor
   endfor
   retval = r ;

endfunction

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% вычисляем изменение весов
% 
% g - градиент функции ошибки
% dW - пред.изменение весов
% W - весовая матрица
% trsp - скорость обучения
% rpar - параметр регуляризации
% mpar - параметр момента
%
function retval = weight_change(g,dW,W,trsp,rpar,mpar),
   s=length(W);
   for i=1:s,
	  % коррекция весовых матриц слоёв с поправками момента и регуляризации
      dW{i}{1} = ( g{i}{1} + rpar*W{i}{1} ) .* trsp + mpar*dW{i}{1} ;
      dW{i}{2} = g{i}{2} .* trsp + mpar*dW{i}{2} ;
   endfor
   retval=dW;
endfunction

% function retval = weight_change(g,W,trsp,rpar),
%    r={};
%    for i=1:length(g),
% 	  r{i}{1} =  ( g{i}{1} + rpar*W{i}{1} ) .* trsp ;
%       r{i}{2} = g{i}{2} .* trsp ;
%    endfor
%    retval=r;
% endfunction


