%!octave 
%
%
%       OS : GNU/Linux 4.2.5-1-ARCH 
% COMPILER : GNU Octave, version 4.0.0
%
%   AUTHOR : Evgeny S. Borisov
%
%  Glushkov Institute of Cybernetics
%  National Academy of Sciences of Ukraine
% 
%    http://www.mechanoid.kiev.ua
%  e-mail : nn@mechanoid.kiev.ua
% 
% $Id: mlp_trn_bfgs.m,v 1.2 2016/04/12 13:19:18 mechanoid Exp $


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% обучение ИНС методом BFGS
%
% L={Xl,Yl} - учебный набор
% С={Xc,Yc} - контрольный набор
% W - начальная матрица весов
% cfg - описание параметров сети сети
% trn_par - параметры обучения
%
function [W,er_hist]=mlp_trn_bfgs(L,C,W,cfg,trn_par),
   ltp = 2 ; % номер ячейки cfg с типом ф-ции активации

   min_err = trn_par{1} ; % порог допустимой ошибки обучения
   max_iter = trn_par{2} ; % максимальное число эпох
   er_win_size = trn_par{3} ;  % размер окна контроля за ошибкой
   trsp = trn_par{4} ; % параметр скорости обучения
   moment_par = trn_par{5} ; % параметр момента 
   reg_par = trn_par{6} ; % параметр регуляризации
   print_step = trn_par{8} ;% печатаем значение ошибки каждые print_step эпох 

   %er_win_size = 5; % размер окна контроля за ошибкой
   %trsp=8e-1;
   %reg_par = 1e-5 ; % параметр регуляризации
   %moment_par=5e-2; % параметр момента 


   tname="BFGS";
   printf("[i] обучение методом %s\n",tname);
   save("-text","result/tname.txt", "tname" ) ;
	  
   
   printf("[i] размер учебного набора (full-batch) %i \n",num_examples);



   er_hist = [er_win_size:-1:1]*100.0 ; % история значений ошибки обучения

   er_hist(end+1) = mlp_loss(C,W,cfg,0.0); % вычисляем ошибку сети на контрольном наборе
   printf("[i]   %i/%i : %f / %f     \r",0,max_iter,er_hist(end),min_err) ; 
   if trn_stop_condition(er_hist,min_err,er_win_size), er_hist(1:er_win_size)=[]; return ; endif % условия остановки обучения

   % первый шаг по методу градиентного спуска
   g = mlp_backprop_grad(L,W,cfg); % значение градиента
   g=mlp_weight_norm(g); 
   dW = mlp_weight_set(cfg,0.0) ; 
   dW = mlp_weight_add(dW,g,trsp); % изменение весов
   W = mlp_weight_add(W,dW,-1.0);% корректируем веса 

   s = mlp_weight_unpack(dW) ; % разворачиваем изменение весов в вектор для вычисления обратного гессиана ф-ции ошибки

   % начальное значение обратного гессиана
   sz_v = size(s,1) * [ 1 1 ];
   V=eye(sz_v); 

   for iter=1:max_iter, % цикл обучения

      g0 = g; % сохраняем старое значение градиента
	  g = mlp_backprop_grad(L,W,cfg); % новое значение градиента
	  g=mlp_weight_norm(g); 
	  gv = mlp_weight_unpack(g); % разворачиваем значение градиента в вектор
	  r = gv - mlp_weight_unpack(g0) ; % изменение градиента ф-ции ошибки

      V = calc_inv_hessian(V,r,s) ; % приближенное значение обратного гессиана ф-ции ошибки 

	  % новое изменение весов
	  s = V * gv;
	  s = s / norm(s);
	  dW0 = mlp_weight_pack(s,cfg);
  
	  dW = weight_change(dW0,dW,W,trsp,reg_par,moment_par); % изменение весов 
	 
	  W = mlp_weight_add(W,dW,-1.0); % корректируем веса 

	  er_hist(end+1) = mlp_loss(C,W,cfg,0.0); % вычисляем ошибку сети на контрольном наборе

	  %print_step=20;
	  if mod(iter,print_step)<1, % печатаем значение ошибки каждые print_step эпох 
		 printf("      %i/%i : %f / %f        \r",iter,max_iter,er_hist(end),min_err) ; 
      endif

	  if trn_stop_condition(er_hist,min_err,er_win_size), er_hist(1:er_win_size)=[]; break ; endif % условия остановки обучения

   endfor

   printf("[i]   %i/%i : %f / %f       \n",iter,max_iter,er_hist(end),min_err) ; 

   er_hist(1:er_win_size)=[]; % удаляем начальные значения ошибки из истории 

   if iter>(max_iter-1) , printf("\n[w] достигнут порог количества циклов обучения         \n") ; endif

endfunction

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval = calc_inv_hessian(V,r,s),
   retval = V - (V*s*s'*V)/(s'*V*s) + (r*r')/(s'*s);
endfunction


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% вычисляем изменение весов
%
function retval = weight_change(g,dW,W,trsp,rpar,mpar),
   s=length(W);
   for i=1:s, % коррекция весовых матриц слоёв с поправками момента и регуляризации
      dW{i}{1} = trsp*( g{i}{1} + rpar*W{i}{1} ) + mpar*dW{i}{1} ;
      dW{i}{2} = trsp*g{i}{2} + mpar*dW{i}{2} ;
   endfor
   retval=dW;
endfunction

