%!octave 
%
%
%       OS : GNU/Linux 4.2.5-1-ARCH 
% COMPILER : GNU Octave, version 4.0.0
%
%   AUTHOR : Evgeny S. Borisov
%
%  Glushkov Institute of Cybernetics
%  National Academy of Sciences of Ukraine
% 
%    http://www.mechanoid.kiev.ua
%  e-mail : nn@mechanoid.kiev.ua
% 
% $Id: mlp_trn_lm.m,v 1.2 2016/04/12 13:19:18 mechanoid Exp $


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% обучение ИНС методом Левенберга-Марквардта
%
% L={Xl,Yl} - учебный набор
% С={Xc,Yc} - контрольный набор
% W - начальная матрица весов
% cfg - описание параметров сети сети
% trn_par - параметры обучения
%
function [W,er_hist]=mlp_trn_lm(L,C,W,cfg,trn_par),
   ltp = 2 ; % номер ячейки cfg с типом ф-ции активации

   min_err = trn_par{1} ; % порог допустимой ошибки обучения
   max_iter = trn_par{2} ; % максимальное число эпох
   er_win_size = trn_par{3} ;  % размер окна контроля за ошибкой
   trsp = trn_par{4} ; % параметр скорости обучения
   moment_par = trn_par{5} ; % параметр момента 
   reg_par = trn_par{6} ; % параметр регуляризации
   batch_par = trn_par{7} ; % доля учебных примеров в одном пакете от общего количества примеров
   print_step = trn_par{8} ; % доля учебных примеров в одном пакете от общего количества примеров

   num_examples=size(L{2},2) ; % количество объектов в учебном наборе
   %batch_size=round(num_examples*batch_par); % количество учебных примеров в одном пакете 
   batch_size = max(2,round(num_examples*batch_par)) ; % количество учебных примеров в одном пакете 


   %er_win_size = 5 ; % размер окна контроля за ошибкой
   %trsp = 1.0 ; % скорость обучения (изменения весов)
   %reg_par = 1e-5 ; % параметр регуляризации
   %reg_par = 0 ; % параметр регуляризации
   %moment_par = 5e-3; % параметр момента 
   
   max_iter_mu = 5;
   mu_fact = 10 ;
   mu = 1e-2;


   tname="Левенберг-Марквардт";
   printf("[i] обучение методом %s\n",tname);
   save("-text","result/tname.txt", "tname" ) ;

   if batch_size < num_examples,
	  printf("[i] размер пакета (mini-batch) %i / %i \n",batch_size,num_examples);
   else
	  printf("[i] размер учебного набора (full-batch) %i \n",num_examples);
   endif


   er_hist = [er_win_size:-1:1]*100.0 ; % история значений ошибки обучения

   % проверяем начальное состояние ошибки сети
   er_hist(end+1) = mlp_loss(C,W,cfg,0.0); % вычисляем ошибку сети на контрольном наборе
   printf("[i]   %i/%i : %f / %f     \r",0,max_iter,er_hist(end),min_err) ; 
   if trn_stop_condition(er_hist,min_err,er_win_size), er_hist(1:er_win_size)=[]; return ; endif % условия остановки обучения

   dW = mlp_weight_set(cfg,0.0) ; % начальные значения изменений весов

   for iter=1:max_iter, % цикл обучения

      % случайный пакет из учебных примеров
	  batch_idx = randperm(num_examples)(1:batch_size); 
	  Lr = { L{1}(:,batch_idx) , L{2}(:,batch_idx) };

	  l = mlp_layers(Lr{1}, W, cfg ) ; % состояния слоёв для всех примеров 
	  o = mlp_activate( l{end}, cfg{end}{ltp} ) ; % выходы сети для всех примеров 
	  oer = ( Lr{2} - o )(:) ; % значения ошибки сети для всех примеров 

	  jcb = calc_jacobian(l,W,cfg) ; % якобиан, матрица частных производных ф-ции ошибки сети на всех примерах

	  % элементы приближенного значения гессиана
	  jcb2 = jcb'*jcb ;
	  djcb2 = eye(size(jcb2)) .* jcb2 ;

	  % значение градиента функции ошибки
	  grd = jcb' * oer ;

	  dW0 = dW ; % сохраняем старые значения изменения весов

	  for i=1:max_iter_mu, % цикл поиска оптимального параметра mu

		 % вычисляем изменение весов
		 dW = inv( jcb2 + djcb2*mu ) * grd ; 
		 dW = dW/norm(dW) ;
		 
		 dW = mlp_weight_pack(dW,cfg); % пакуем вектор изменения весов в матрицы
		 
		 dW = weight_change(dW,dW0,W,trsp,reg_par,moment_par); % изменение весов 

		 Wnew = mlp_weight_add(W,dW,-1.0); % корректируем веса 

		 er = mlp_loss(C,Wnew,cfg,0.0); % вычисляем ошибку сети на контрольном наборе

		 if er < er_hist(end), % ошибка уменьшилась
			mu/=mu_fact ;  % завершаем цикл поиска mu
			break;
		 endif

		 % ошибка выросла
		 mu*=mu_fact ; % изменяем параметр и проверяем изменение ошибки
	  endfor

	  W=Wnew; % сохраняем результат и переходим к следующей эпохе

	  er_hist(end+1) = er ; % сохраняем текущее значение ошибки сети
	  if mod(iter,print_step)<1, % печатаем значение ошибки каждые print_step эпох 
		 printf("      %i/%i : %f / %f  ,   %f     \r",iter,max_iter,er_hist(end),min_err,mu) ; 
      endif


	  if trn_stop_condition(er_hist,min_err,er_win_size), break ; endif % условия остановки обучения

   endfor

   er_hist(1:er_win_size)=[]; % удаляем начальные значения ошибки из истории 

   if iter>(max_iter-1) , printf("\n[w] достигнут порог количества циклов обучения         \n") ; endif

endfunction



% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% якобиан, матрица частных производных ф-ции ошибки сети на всех примерах
function retval = calc_jacobian(l,W,cfg),
   layers_numb = length(cfg) ; % количество слоёв сети, включая вход

   % элементы якобиана для весов последнего слоя
   d = calc_delta_layer_last(l,W,cfg); 
   [ Jcb_n Jcb_b ] = calc_jacobian_layer(l,W,cfg,d,layers_numb); 
 
   % элементы якобиана для скрытых слоёв
   for i=layers_numb-1:-1:2,
	  d = calc_delta_layer(l,W,cfg,d,i); 
	  [ Jcb_n_i Jcb_b_i ] = calc_jacobian_layer(l,W,cfg,d,i); 
	  Jcb_n = [ Jcb_n_i Jcb_n ]; % элементы якобиана для весов
	  Jcb_b = [ Jcb_b_i Jcb_b ]; % элементы якобиана для сдвигов
   endfor

   retval = [ Jcb_n Jcb_b ]  ;
endfunction


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function [ jcb_n jcb_b ] = calc_jacobian_layer(l,W,cfg,d,n),
   ltp = 2 ; % номер ячейки cfg с типом ф-ции активации

   sz_layer_out = size(l{end},1) ; % размер выходного слоя
   sz_layer_n   = size(l{n},1) ;   % размер слоя n
   sz_layer_n0  = size(l{n-1},1) ; % размер слоя n-1
   ex_numb      = size(l{end},2) ; % количество примеров

   % значение входов нейронов предпоследнего слоя (т.е.выходов предпоследнего слоя) для всех примеров
   x = mlp_activate( l{n-1} , cfg{n-1}{ltp} ) ; 
    
   % jcb=d*x
   jcb_n = zeros( [ (ex_numb*sz_layer_out) , (sz_layer_n*(sz_layer_n0)) ] ); % элементы якобиана для весов
   jcb_b = zeros( [ (ex_numb*sz_layer_out) , sz_layer_n ] ); % элементы якобиана для сдвигов

   for p=1:ex_numb, % для всех примеров
	  for j=1:sz_layer_out, % для всех выходов сети
		 dej = d(j,:,p)'*x(:,p)' ; % елементы якобиана для весов слоя n, выхода j, на примере p
		 k=(p-1)*sz_layer_out+j; % номер строки части якобиана для слоя n
         jcb_n(k,:)=dej(:)' ; % разворачиваем в вектор
         jcb_b(k,:)=d(j,:,p) ; 
	  endfor 
   endfor

endfunction

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval = calc_delta_layer(l,W,cfg,d0,n),
   ltp = 2 ; % номер ячейки cfg с типом ф-ции активации

   sz_layer_out = size(l{end},1) ; % размер выходного слоя
   sz_layer_n   = size(l{n},1) ;   % размер слоя n
   ex_numb      = size(l{end},2) ; % количество примеров

   % значение производной ф-ции активации на предпоследнем слое для всех примеров 
   df = mlp_activate_drvt( l{n} , cfg{n}{ltp} ) ; 

   d=zeros( [sz_layer_out, sz_layer_n, ex_numb] );
   for p = 1:ex_numb, % для всех примеров 
	   d(:,:,p) =  ( d0(:,:,p) * W{n}{1} ) .* df(:,p)' ; % значения delta для примера p
   endfor

   retval = d ; % ex_numb матриц размера [ sz_layer_out  x sz_layer_n ]
endfunction


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval = calc_delta_layer_last(l,W,cfg),
   ltp = 2 ; % номер ячейки cfg с типом ф-ции активации

   sz_layer_out = size(l{end},1) ; % размер выходного слоя
   ex_numb      = size(l{end},2) ; % количество примеров

   % значение производной ф-ции активации на последнем слое для всех примеров (sz_layer_out x ex_numb)
   df = mlp_activate_drvt( l{end} , cfg{end}{ltp} ) ; 

   % для производной активации softmax
   if ndims(df) > 2, retval=-df; return; endif
   

   % для производных прочих функций активации 
   df3 = permute(reshape(df',[1 ex_numb sz_layer_out]) , [ 3 1 2] ) ; % разбираем матрицу на столбцы [sz_layer_out 1 ex_numb]
   d = ones([1,1,ex_numb]).*eye([sz_layer_out,sz_layer_out]) ; % ex_numb единичных матриц размера [ sz_layer_out sz_layer_out ]

   retval = - d.*df3 ; % ex_numb матриц размера [ sz_layer_out sz_layer_out ]
      
endfunction

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% вычисляем изменение весов
% 
function retval = weight_change(g,dW,W,trsp,rpar,mpar),
   s=length(W);
   for i=1:s,
	  % коррекция весовых матриц слоёв с поправками момента и регуляризации
      dW{i}{1} = trsp*( g{i}{1} + rpar*W{i}{1} ) + mpar*dW{i}{1} ;
      dW{i}{2} = trsp*g{i}{2} + mpar*dW{i}{2} ;
   endfor
   retval=dW;
endfunction

