%!octave 
%
%       OS : GNU/Linux 4.2.5-1-ARCH 
% COMPILER : GNU Octave, version 4.0.0
%
%   AUTHOR : Evgeny S. Borisov
%
%  Glushkov Institute of Cybernetics
%  National Academy of Sciences of Ukraine
% 
%    http://www.mechanoid.kiev.ua
%  e-mail : nn@mechanoid.kiev.ua
% 
% $Id: mlp_trn_qprop.m,v 1.2 2016/04/12 13:19:18 mechanoid Exp $


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% обучение ИНС методом quickProp
%
% L={Xl,Yl} - учебный набор
% С={Xc,Yc} - контрольный набор
% W - начальная матрица весов
% cfg - описание параметров сети сети
% trn_par - параметры обучения
%
function [W,er_hist]=mlp_trn_qprop(L,C,W,cfg,trn_par),
   ltp = 2 ; % номер ячейки cfg с типом ф-ции активации

   min_err = trn_par{1} ; % порог допустимой ошибки обучения
   max_iter = trn_par{2} ; % максимальное число эпох
   er_win_size = trn_par{3} ;  % размер окна контроля за ошибкой
   trsp0 = trn_par{4} ; % параметр скорости обучения
   %moment_par = trn_par{5} ; % параметр момента 
   reg_par = trn_par{6} ; % параметр регуляризации
   batch_par = trn_par{7} ; % доля учебных примеров в одном пакете от общего количества примеров
   print_step = trn_par{8} ;% печатаем значение ошибки каждые print_step эпох 


   num_examples=size(L{2},2) ; % количество объектов в учебном наборе
   % batch_size=round(num_examples*batch_par); % количество учебных примеров в одном пакете 
   batch_size = max(2,round(num_examples*batch_par)) ; % количество учебных примеров в одном пакете 

   tname="quickProp";
   printf("[i] обучение методом %s\n",tname);
   save("-text","result/tname.txt", "tname" ) ;

   if batch_size < num_examples,
	  printf("[i] размер пакета (mini-batch) %i / %i \n",batch_size,num_examples);
   else
	  printf("[i] размер учебного набора (full-batch) %i \n",num_examples);
   endif



 
   er_hist = [er_win_size:-1:1]*100.0 ; % история значений ошибки обучения

   % случайный пакет из учебных примеров
   batch_idx = randperm(num_examples)(1:batch_size); 
   Lr = { L{1}(:,batch_idx) , L{2}(:,batch_idx) };

   g = mlp_backprop_grad(Lr,W,cfg); % значение градиента
   g = mlp_weight_norm(g); 

   dW = mlp_weight_set(cfg,0.0) ; % начальные значения изменения весов
   trsp = mlp_weight_set(cfg,trsp0) ; % начальные значения  скорости обучения
   moment_par = mlp_weight_set(cfg,0.0) ; % начальные значения параметра момента

   dW = weight_change_qprop(g,dW,W,trsp,reg_par,moment_par);% изменение весов 
   
   W0=W;
   W = mlp_weight_add(W,dW,-1.0);% корректируем веса 

   for iter=1:max_iter,
	  er_hist(end+1) = mlp_loss(C,W,cfg,reg_par); % вычисляем ошибку сети на контрольном наборе

	  if mod(iter,print_step)<1, % печатаем значение ошибки print_step эпох 
		 printf("      %i/%i : %f / %f        \r",iter,max_iter,er_hist(end),min_err) ; 
      endif
 	  
	  if trn_stop_condition(er_hist,min_err,er_win_size), break ; endif % условия остановки обучения

	  g0 = g; % сохраняем старое значение градиента ф-ции потери

	  % случайный пакет из учебных примеров
	  batch_idx = randperm(num_examples)(1:batch_size); 
	  Lr = { L{1}(:,batch_idx) , L{2}(:,batch_idx) };

	  g = mlp_backprop_grad(Lr,W,cfg); % новое значение градиента ф-ции потери
	  g = mlp_weight_norm(g); 

	  moment_par = get_moment_param(g,W,g0,W0,reg_par); % параметры момента для каждого веса

	  trsp = trn_speed_param(g,dW,W,reg_par,trsp0); % скорость обучения для каждого веса

	  dW = weight_change_qprop(g,dW,W,trsp,reg_par,moment_par);% изменение весов 

	  W0 = W ; % сохраняем старые значения весов
	  W = mlp_weight_add(W,dW,-1.0);% корректируем веса 
   endfor

   printf("[i]   %i/%i : %f / %f        \n",iter,max_iter,er_hist(end),min_err) ; 

   er_hist(1:er_win_size)=[]; % удаляем начальные значения ошибки из истории 

   if iter>(max_iter-1) , printf("[w] достигнут порог количества циклов обучения         \n") ; endif

endfunction


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% скорость обучения для каждого веса
%
function retval= trn_speed_param(g,dW,W,rpar,trsp0),
   trsp={};
   s=length(W);
   for i=1:s,
	  md=( ((g{i}{1}+W{i}{1}.*rpar).*-dW{i}{1}) > 0.0 );
      mz=(dW{i}{1}!=0.0);
	  m=md|mz;
	  trsp{i}{1}=m.*trsp0;

	  md=( (g{i}{2}.*-dW{i}{2}) > 0.0 );
      mz=(dW{i}{2}!=0.0);
	  m=md|mz;
	  trsp{i}{2}=m.*trsp0;
   endfor

   retval=trsp;
endfunction


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% изменение весов
%
function retval= weight_change_qprop(g,dW,W,trsp,rpar,mpar),
   s=length(W);
   for i=1:s,
	  % коррекция весовых матриц слоёв с поправками момента и регуляризации
      dW{i}{1} = trsp{i}{1}.*( g{i}{1} + rpar*W{i}{1} ) + mpar{i}{1}.*dW{i}{1} ;
      dW{i}{2} = trsp{i}{2}.*g{i}{2} + mpar{i}{2}.*dW{i}{2} ;
   endfor
   retval=dW;

endfunction

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% параметры момента для каждого веса
%
function retval= get_moment_param(g,W,g0,W0,rpar),
   %mpar=1e-3;
   a_max=1.75;

   S  = mlp_weight_add( g, W, rpar); % S=g+rpar*W;
   S0 = mlp_weight_add(g0, W0,rpar); 

   dS = mlp_weight_add(S0, S, -1);   % ds = s0 - s
   dW = mlp_weight_add(W0, W, -1); 

   B = weight_div(S,dS); % B=S./(S_old-S);

   D = mlp_weight_mult(S,dW); % (S.*dW.*B)
   D = mlp_weight_mult(D,B);

   A=weight_min(B,a_max,D); 

   retval=A;
endfunction

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% if d<0 or b>a then res=a else res=b endif
%
function retval = weight_min(b,a,d),
   s=length(b);
   r={};
   for i=1:s,
	  for j=1:2,
		 m = ( d{i}{j} < 0.0 ) | ( b{i}{j} > a );
		 r{i}{j}=b{i}{j}.*not(m)+(m.*a);
	  endfor
   endfor
   retval=r;
endfunction

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval = weight_div(w1,w2),
   s=length(w1);
   r={};
   for i=1:s,
	  for j=1:2,
		 %  если знаменатель нулевой то определяем результат деления = 0
	     d=(w2{i}{j}!=0.0); 
		 z=w2{i}{j}+not(d);
		 r{i}{j} = (w1{i}{j} ./ z).*d ; 
	  endfor
   endfor
   retval=r;
endfunction
 


