%!octave 
%
%
%       OS : GNU/Linux 4.2.5-1-ARCH 
% COMPILER : GNU Octave, version 4.0.0
%
%   AUTHOR : Evgeny S. Borisov
%
%  Glushkov Institute of Cybernetics
%  National Academy of Sciences of Ukraine
% 
%    http://www.mechanoid.kiev.ua
%  e-mail : nn@mechanoid.kiev.ua
% 
% $Id: mlp_trn_rprop.m,v 1.2 2016/04/12 13:19:18 mechanoid Exp $


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% обучение ИНС методом rProp (resilient back propagation)
%
% L={Xl,Yl} - учебный набор
% С={Xc,Yc} - контрольный набор
% W - начальная матрица весов
% cfg - описание параметров сети сети
% trn_par - параметры обучения
%
function [W,er_hist]=mlp_trn_rprop(L,C,W,cfg,trn_par),
   ltp = 2 ; % номер ячейки cfg с типом ф-ции активации

   min_err = trn_par{1} ; % порог допустимой ошибки обучения
   max_iter = trn_par{2} ; % максимальное число эпох
   er_win_size = trn_par{3} ;  % размер окна контроля за ошибкой
   trsp0 = trn_par{4} ; % параметр скорости обучения
   moment_par = trn_par{5} ; % параметр момента 
   reg_par = trn_par{6} ; % параметр регуляризации
   batch_par = trn_par{7} ; % доля учебных примеров в одном пакете от общего количества примеров
   print_step = trn_par{8} ; % печатаем значение ошибки print_step эпох

   %er_win_size = 5; % размер окна контроля за ошибкой
   %trsp0 = 5e-3; % скорость обучения (изменения весов)
   %reg_par = 1e-4 ; % параметр регуляризации
   %moment_par = 5e-3; % параметр момента 


   num_examples=size(L{2},2) ; % количество объектов в учебном наборе
   %batch_size=round(num_examples*batch_par); % количество учебных примеров в одном пакете 
   batch_size = max(2,round(num_examples*batch_par)) ; % количество учебных примеров в одном пакете 

   tname="rProp";
   printf("[i] обучение методом %s\n",tname);
   save("-text","result/tname.txt", "tname" ) ;

   if batch_size < num_examples,
	  printf("[i] размер пакета (mini-batch) %i / %i \n",batch_size,num_examples);
   else
	  printf("[i] размер учебного набора (full-batch) %i \n",num_examples);
   endif


   er_hist = [er_win_size:-1:1]*100.0 ; % история значений ошибки обучения

   % случайный пакет из учебных примеров
   batch_idx = randperm(num_examples)(1:batch_size); 
   Lr = { L{1}(:,batch_idx) , L{2}(:,batch_idx) };

   g = mlp_backprop_grad(Lr,W,cfg); % значение градиента
   g = mlp_weight_norm(g); 

   dW = mlp_weight_set(cfg,0.0) ; 
   dW = mlp_weight_add(dW,g,trsp0); % изменение весов
   W = mlp_weight_add(W,dW,-1.0);% корректируем веса 

   trsp = mlp_weight_set(cfg,trsp0) ; % начальные значения скорости обучения для каждого параметра

   for iter=1:max_iter,

	  er_hist(end+1) = mlp_loss(C,W,cfg,0.0); % вычисляем ошибку сети на контрольном наборе

	  if mod(iter,print_step)<1, % печатаем значение ошибки print_step эпох 
		 printf("      %i/%i : %f / %f     \r",iter,max_iter,er_hist(end),min_err) ; 
      endif
	  if trn_stop_condition(er_hist,min_err,er_win_size), break ; endif % условия остановки обучения

	  g0 = g; % сохраняем старое значение градиента 

	  % случайный пакет из учебных примеров
	  batch_idx = randperm(num_examples)(1:batch_size); 
	  Lr = { L{1}(:,batch_idx) , L{2}(:,batch_idx) };

	  g = mlp_backprop_grad(Lr,W,cfg); % новое значение градиента
	  g = mlp_weight_norm(g); 

	  trsp = get_trsp_param(g,g0,trsp); % вычисляем параметр скорости обучения для каждого веса сети

	  dW = weight_change(g,W,dW,trsp,reg_par,moment_par);% изменение весов 
	  W = mlp_weight_add(W,dW,-1.0);% корректируем веса 

   endfor

   printf("[i]   %i/%i : %f / %f     \n",iter,max_iter,er_hist(end),min_err) ; 

   er_hist(1:er_win_size)=[]; % удаляем начальные значения ошибки из истории 

   if iter>(max_iter-1) , printf("[w] достигнут порог количества циклов обучения         \n") ; endif

endfunction


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval= get_trsp_param(g,g0,trsp),
   a=1.2;
   b=0.5;
   t_min=1e-5;
   t_max=20.0;

   S = mlp_weight_mult(g,g0);
   R = trsp; % S == 0.0
   R = get_trsp_param_min(S,R,a,t_max);
   R = get_trsp_param_max(S,R,b,t_min);
   
   retval = R;

endfunction

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% dW=-trsp*sign(g);
function retval= weight_change(g,W,dW,trsp,rpar,mpar),

   s=length(g);
   R=trsp;
   for i=1:s,
	  R{i}{1} = trsp{i}{1}.*( sign(g{i}{1}) + rpar*W{i}{1} ) + mpar*dW{i}{1}  ; 
	  R{i}{2} = trsp{i}{2}.*sign(g{i}{2}) + mpar*dW{i}{2}  ; 
   endfor
   retval=R;
endfunction



% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% S>0.0 -> min(a*trsp,t_max)
function retval= get_trsp_param_min(S,trsp,a,t_max),
   s=length(S);
   R=trsp;
   for i=1:s,
	  for j=1:2,
		 m = ( S{i}{j} >0.0);
		 t=min( a * trsp{i}{j}, t_max ) ;
		 R{i}{j} = trsp{i}{j}.*not(m) + t .* m ; 
	  endfor
   endfor
   retval=R;
endfunction

% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% S<0.0 -> max(b*trsp,t_min)
function retval= get_trsp_param_max(S,trsp,b,t_min),
   s=length(S);
   R=trsp;
   for i=1:s,
	  for j=1:2,
		 m = (S{i}{j}<0.0);
		 t=max( b * trsp{i}{j}, t_min ) ;
		 R{i}{j} = trsp{i}{j}.*not(m) + t .* m ; 
	  endfor
   endfor
   retval=R;
endfunction




