%!octave 
%
%
%       OS : GNU/Linux 4.2.5-1-ARCH 
% COMPILER : GNU Octave, version 4.0.0
%
%   AUTHOR : Evgeny S. Borisov
%
%  Glushkov Institute of Cybernetics
%  National Academy of Sciences of Ukraine
% 
%    http://www.mechanoid.kiev.ua
%  e-mail : nn@mechanoid.kiev.ua
% 
% $Id: mlp_trn_sgd.m,v 1.1 2016/04/20 08:18:24 mechanoid Exp $


% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
% обучение ИНС методом градиентного спуска (стохастический)
%
% L={Xl,Yl} - учебный набор
% С={Xc,Yc} - контрольный набор
% W - начальная матрица весов
% cfg - описание параметров сети сети
% trn_par - параметры обучения
%
function [W,er_hist]=mlp_trn_sgd(L,C,W,cfg,trn_par),
   ltp = 2 ; % номер ячейки cfg с типом ф-ции активации

   min_err = trn_par{1} ; % порог допустимой ошибки обучения
   max_iter = trn_par{2} ; % максимальное число эпох
   er_win_size = trn_par{3} ;  % размер окна контроля за ошибкой
   trsp = trn_par{4} ; % параметр скорости обучения
   %moment_par = trn_par{5} ; % параметр момента 
   %reg_par = trn_par{6} ; % параметр регуляризации
   batch_par = trn_par{7} ; % доля учебных примеров в одном пакете от общего количества примеров
   print_step = trn_par{8} ;% печатаем значение ошибки каждые print_step эпох 

   %batch_par=0.7; % доля учебных примеров в одном пакете от общего количества примеров
   %er_win_size=5; % размер окна контроля за ошибкой
   %trsp=1e-1; % скорость обучения (изменения весов)
   %trsp=3e-2; % скорость обучения (изменения весов)
   %reg_par=1e-5 ; % параметр регуляризации
   %reg_par = 2.0 / num_examples  % параметр регуляризации
   %moment_par=5e-2; % параметр момента 

   X=L{1} ; y=L{2} ; % учебный набор
   num_examples=size(y,2) ; % количество объектов в учебном наборе
   batch_size = max(2,round(num_examples*batch_par)) ; % количество учебных примеров в одном пакете 
   
   tname="градиентный спуск (стохастический)";
   printf("[i] обучение методом %s\n",tname);
   save("-text","result/tname.txt", "tname" ) ;

   printf("[i] размер учебного набора  %i \n",num_examples);
   printf("[i] проверка значения ошибки через каждые %i примеров \n",batch_size);

 
   er_hist = [er_win_size:-1:1]*100.0 ; % история значений ошибки обучения

   W0=W;

   for iter=1:max_iter,

	  er_hist(end+1) = mlp_loss(C,W,cfg,0.0); % вычисляем ошибку сети на контрольном наборе

	  if mod(iter,print_step)<1, % печатаем значение ошибки print_step эпох 
		 printf("      %i/%i : %f / %f   , %f     \r",iter,max_iter,er_hist(end),min_err,trsp ) ; 
      endif
 	 
	  if trn_stop_condition(er_hist,min_err,er_win_size), break ; endif % условия остановки обучения
 
	  % batch_par=0.01 ;
	  de = er_hist(end) - er_hist(end-1) ; % изменение ошибки
 	  if( de > 0 ), % ошибка выросла  
		 p=1.0/(1.0+exp(-de)); % вероятность возврата старого состояния в случае увеличения ошибки
		 if (rand() < p), 
		    %printf("\n[i] ошибка выросла, возврат старого состояния с вероятностью %f\n",p);
			W=W0; 
			er_hist(end)=[];
		 endif
 	  endif
   
	  W0=W;

	  % случайный пакет из учебных примеров
      batch_idx = randperm(num_examples)(1:batch_size); 
      
	  for iter2=batch_idx,
		 Lr = { X(:,iter2) , y(:,iter2) };
		 g = mlp_backprop_grad(Lr,W,cfg); % значение градиента
		 % g=mlp_weight_norm(g); 
		 W = mlp_weight_add(W,g,-trsp); % корректируем веса 
	  endfor
	 
      %trsp=trn_speed(trsp,er_hist); % скорость обучения (изменения весов)

   endfor

   printf("[i]   %i/%i : %f / %f   , %f     \n",iter,max_iter,er_hist(end),min_err,trsp ) ; 

   er_hist(1:er_win_size)=[]; % удаляем начальные значения ошибки из истории 

   if iter>(max_iter-1) , printf("[w] достигнут порог количества циклов обучения         \n") ; endif

endfunction



% - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
function retval=trn_speed(s,er),
   k=1.01; % коэффициент допустимого роста ошибки
   ri=1.01; % коэффициент увеличения скорости
   rd=0.99; % коэффициент уменьшения скорости

   d = er(end) > k*er(end-1) ;

   if( d > 0 ),  % если существенный рост ошибки
	  s*=rd; % то уменьшаем скорость обучения
   else 
	  s=min(s*ri,1.0); % иначе увеличиваем скорость обучения
   endif
   retval=s ;
endfunction


